package server

import (
	"errors"
	"log"
	"regexp"
	"time"

	ksclient "git.autistici.org/id/keystore/client"
	"github.com/gorilla/securecookie"

	"git.autistici.org/id/go-sso/server/device"
)

// Config data for the SSO service.
type Config struct {
	SecretKeyFile    string   `yaml:"secret_key_file"`
	PublicKeyFile    string   `yaml:"public_key_file"`
	Domain           string   `yaml:"domain"`
	AllowedServices  []string `yaml:"allowed_services"`
	AllowedExchanges []*struct {
		SrcRegexp string `yaml:"src_regexp"`
		DstRegexp string `yaml:"dst_regexp"`
		srcRx     *regexp.Regexp
		dstRx     *regexp.Regexp
	} `yaml:"allowed_exchanges"`
	ServiceTTLs []*struct {
		Regexp     string `yaml:"regexp"`
		TTLSeconds int    `yaml:"ttl"`
		rx         *regexp.Regexp
	} `yaml:"service_ttls"`
	AuthSessionLifetimeSeconds int              `yaml:"auth_session_lifetime"`
	SessionSecrets             []string         `yaml:"session_secrets"`
	CSRFSecret                 string           `yaml:"csrf_secret"`
	AuthService                string           `yaml:"auth_service"`
	DeviceManager              *device.Config   `yaml:"device_manager"`
	KeyStore                   *ksclient.Config `yaml:"keystore"`

	allowedServicesRx []*regexp.Regexp
}

// Check syntax (missing required values).
func (c *Config) valid() error {
	if c.SecretKeyFile == "" {
		return errors.New("secret_key_file is empty")
	}
	if c.PublicKeyFile == "" {
		return errors.New("public_key_file is empty")
	}
	if c.Domain == "" {
		return errors.New("domain is empty")
	}
	if len(c.AllowedServices) == 0 {
		return errors.New("the list of allowed services is empty")
	}
	if c.AuthService == "" {
		return errors.New("auth_service is empty")
	}

	// Some things we can autogenerate, but for testing purposes
	// only. Print a warning.
	if len(c.SessionSecrets) == 0 {
		log.Printf("Warning: session_secrets unset, generating temporary random session secrets")
		c.SessionSecrets = []string{
			string(securecookie.GenerateRandomKey(64)),
			string(securecookie.GenerateRandomKey(32)),
		}
	}

	return nil
}

// Compile the configuration (parse regular expressions, etc).
func (c *Config) Compile() error {
	err := c.valid()
	if err != nil {
		return err
	}

	for _, svcttl := range c.ServiceTTLs {
		svcttl.rx, err = regexp.Compile(svcttl.Regexp)
		if err != nil {
			return err
		}
	}
	for _, xch := range c.AllowedExchanges {
		xch.srcRx, err = regexp.Compile(xch.SrcRegexp)
		if err != nil {
			return err
		}
		xch.dstRx, err = regexp.Compile(xch.DstRegexp)
		if err != nil {
			return err
		}
	}

	for _, pattern := range c.AllowedServices {
		rx, err := regexp.Compile(pattern)
		if err != nil {
			return err
		}
		c.allowedServicesRx = append(c.allowedServicesRx, rx)
	}

	return nil
}

var defaultServiceTTL = 300 * time.Second

func (c *Config) getServiceTTL(service string) time.Duration {
	for _, svcttl := range c.ServiceTTLs {
		if svcttl.rx.MatchString(service) {
			return time.Duration(svcttl.TTLSeconds) * time.Second
		}
	}
	return defaultServiceTTL
}

func (c *Config) isServiceAllowed(service string) bool {
	for _, rx := range c.allowedServicesRx {
		if rx.MatchString(service) {
			return true
		}
	}
	return false
}

func (c *Config) isExchangeAllowed(src, dst string) bool {
	for _, xch := range c.AllowedExchanges {
		if xch.srcRx.MatchString(src) && xch.dstRx.MatchString(dst) {
			return true
		}
	}
	return false
}
